#ifndef VORONOTALT_VORONOTALT_H_
#define VORONOTALT_VORONOTALT_H_

#include "conversion_to_input.h"
#include "radical_tessellation.h"
#include "simplified_aw_tessellation.h"
#include "updateable_radical_tessellation.h"

#endif /* VORONOTALT_VORONOTALT_H_ */
